require "defines"
require "config"

local SLIPSTREAM_VERSION = "2.0.1"
local energy_per_action = 5000

local next = next
local north = defines.direction.north
local east = defines.direction.east
local south = defines.direction.south
local west = defines.direction.west
local transport_line = defines.transport_line

script.on_init(function()
	onLoad()
end)

script.on_load(function()
	onLoad()
end)

function onLoad()
	if not global.version then
		global = {}
		global.sschests = {}
		global.schedule = {}
		global.version = SLIPSTREAM_VERSION
		game.forces.player.reset_technologies()
		game.forces.player.reset_recipes()
	end
	
	-- if global.version ~= SLIPSTREAM_VERSION then
		-- slipstreamMigration(global.version, SLIPSTREAM_VERSION)
	-- end
end

script.on_event(defines.events.on_tick, function(event)
	if type(global.schedule[game.tick]) == "table" then
		for i,v in pairs(global.schedule[game.tick]) do
			local schedule_delay, error = runChestInstructions(global.sschests[v[1]])
			
			if error == false then
				for _,player in pairs(game.players) do
					player.print("Problem with slipstream chest at " .. global.sschests[v[1]].entity.position.x .. ", " .. global.sschests[v[1]].entity.position.y .. ". Deactivating chest.")
				end
				break
			end
			
			if schedule_delay == false then
				scheduleAdd(v[1], game.tick + 10, 10)
			else
				scheduleAdd(v[1], game.tick + v[2], math.min(v[2] + 10, 120))
			end
		end
		
		global.schedule[game.tick] = nil
	end
end)

script.on_event(defines.events.on_built_entity, function(event)
	if string.find(event.created_entity.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamBuilt(event.created_entity)
	end
end)

script.on_event(defines.events.on_robot_built_entity, function(event)
	if string.find(event.created_entity.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamBuilt(event.created_entity)
	end
end)

script.on_event(defines.events.on_player_mined_item, function(event)
	if string.find(event.item_stack.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamRemoved()
	end
end)

script.on_event(defines.events.on_robot_mined, function(event)
	if string.find(event.item_stack.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamRemoved()
	end
end)

script.on_event(defines.events.on_entity_died, function(event)
	if string.find(event.entity.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamRemoved()
	end
end)

function scheduleAdd(entity, tick, delay) -- For adding new entries to the scheduling table. 'entity' is slightly misleading; it's the coordinate pair of the chest as a string.
	if global.schedule[tick] == nil then
		global.schedule[tick] = {}
	end
	
	table.insert(global.schedule[tick],{entity,delay})
end

function slipstreamBuilt(entity) -- Construct the data table for each chest as it's built.
	local chest_name = entity.name
	
	local new_chest = {}
	
	-- Record position of chest and search area.
	local surface = entity.surface
	local pos = {x = entity.position.x, y = entity.position.y}
	local search = {{pos.x - 0.5,pos.y - 0.5},{pos.x + 0.5, pos.y + 0.5}}
	
	entity.destroy() -- Remove initial chest.
	
	-- Creates the accumulator entity that serves as the chest's power supply.
	new_chest.battery = surface.create_entity{name="slipstream-chest-battery", position = pos, force = game.forces.player}
	new_chest.battery.minable = false
	new_chest.battery.destructible = false
	
	-- Recreate the chest on top of the accumulator.
	new_chest.entity = surface.create_entity{name = chest_name, position = pos, force = game.forces.player}
	new_chest.inventory = new_chest.entity.get_inventory(defines.inventory.chest)
	
	-- Autoconfigure output based on dummy's direction and presence of transport belt.
	new_chest.instructions, new_chest.belts = findInputsOutputs(new_chest.entity)
	new_chest.delay = 10
	
	local index = string.format("%d_%d", new_chest.entity.position.x, new_chest.entity.position.y)
	
	global.sschests[index] = new_chest
	scheduleAdd(index, game.tick + 10, 10)
end

function slipstreamRemoved()
	for i,v in pairs(global.sschests) do
		if not v["entity"].valid then
			v.battery.destroy()
			global.sschests[i] = nil
		end
	end
	
	for i,v in pairs(global.schedule) do
		for j,u in ipairs(v) do
			if global.sschests[u[1]] == nil then
				global.schedule[i][j] = nil
			end
		end
	end
end

function findInputsOutputs(chest) -- Find the input and output belts for a chest.
	local pos = {x = chest.position.x, y = chest.position.y}
	
	local beltscan_coords = { -- Points to search for transport belts.
		[north] = {{pos.x - 0.1, pos.y - 1.1},{pos.x + 0.1, pos.y - 0.9}},
		[east] = {{pos.x + 0.9, pos.y - 0.1},{pos.x + 1.1, pos.y + 0.1}},
		[south] = {{pos.x - 0.1, pos.y + 0.9},{pos.x + 0.1, pos.y + 1.1}},
		[west] = {{pos.x - 1.1, pos.y - 0.1},{pos.x - 0.9, pos.y + 0.1}}
	}
	
	local directions = {north, east, south, west} -- For the for loop.
	local away_directions = {[north] = north, [east] = east, [south] = south, [west] = west}
	local facing_directions = {[north] = south, [east] = west, [south] = north, [west] = east}
	local side_clockwise_directions = {[north] = east, [east] = south, [south] = west, [west] = north}
	local side_anticlockwise_directions = {[north] = west, [east] = north, [south] = east, [west] = south}
	local instructions = {}
	local belts = {}
	
	for i,v in ipairs(directions) do
		local belt = chest.surface.find_entities_filtered{area = beltscan_coords[v], type = "transport-belt"} -- Search for a transport belt.
		if belt[1] ~= nil then -- If belt is found.
			if belt[1].direction == away_directions[v] then -- If belt is facing away from the chest, add both its transport lines to spawns_list and +2 to output limit counter.
				table.insert(instructions,{"output", belt[1].get_transport_line(transport_line.left_line)})
				table.insert(instructions,{"output", belt[1].get_transport_line(transport_line.right_line)})
				belts[belt[1]] = 1
			elseif belt[1].direction == side_clockwise_directions[v] then -- Just one lane.
				table.insert(instructions,{"output", belt[1].get_transport_line(transport_line.right_line)})
				belts[belt[1]] = 1
			elseif belt[1].direction == side_anticlockwise_directions[v] then -- Or the other lane.
				table.insert(instructions,{"output", belt[1].get_transport_line(transport_line.left_line)})
				belts[belt[1]] = 1
			elseif belt[1].direction == facing_directions[v] then
				table.insert(instructions,1,{"input", belt[1].get_transport_line(transport_line.right_line)})
				table.insert(instructions,1,{"input", belt[1].get_transport_line(transport_line.left_line)})
				belts[belt[1]] = 1
			end
		end
	end
	
	return instructions, belts
end

function runChestInstructions(chest)
	local instructions = chest.instructions
	local action_count = 0
	
	for i,v in pairs(instructions) do
		if chest.battery.energy < energy_per_action then
			break
		elseif v[2].valid == false then
			return true, false
		end
		
		if v[1] == "output" and not chest.inventory.is_empty() then
			local chest_contents = next(chest.inventory.get_contents())
			local stack = {name = chest_contents, count = 1}
			
			if v[2].can_insert_at_back() then
				v[2].insert_at_back(stack)
				chest.inventory.remove(stack)
				chest.battery.energy = chest.battery.energy - energy_per_action
				action_count = action_count + 1
			end
		elseif v[1] == "input" then
			local line_contents = next(v[2].get_contents())
			
			if line_contents ~= nil then
				local stack = {name = line_contents, count = 1}
				
				if chest.inventory.can_insert(stack) then
					chest.inventory.insert(stack)
					v[2].remove_item(stack)
					chest.battery.energy = chest.battery.energy - energy_per_action
					action_count = action_count + 1
				end
			end
		end
	end
	
	if action_count == 0 then
		return true, true
	else
		return false, true
	end
end